@@ -4,9 +4,11 @@
 
 package binding
 
-import "net/http"
+import (
+	"net/http"
+)
 
-const defaultMemory = 32 * 1024 * 1024
+const defaultMemory = 32 << 20
 
 type formBinding struct{}
 type formPostBinding struct{}
@@ -20,7 +22,11 @@ func (formBinding) Bind(req *http.Request, obj interface{}) error {
 	if err := req.ParseForm(); err != nil {
 		return err
 	}
-	req.ParseMultipartForm(defaultMemory)
+	if err := req.ParseMultipartForm(defaultMemory); err != nil {
+		if err != http.ErrNotMultipart {
+			return err
+		}
+	}
 	if err := mapForm(obj, req.Form); err != nil {
 		return err
 	}
@@ -49,8 +55,9 @@ func (formMultipartBinding) Bind(req *http.Request, obj interface{}) error {
 	if err := req.ParseMultipartForm(defaultMemory); err != nil {
 		return err
 	}
-	if err := mapForm(obj, req.MultipartForm.Value); err != nil {
+	if err := mappingByPtr(obj, (*multipartRequest)(req), "form"); err != nil {
 		return err
 	}
+
 	return validate(obj)
 }
